import { syntaxTree, LRLanguage, LanguageSupport } from '@codemirror/language';
import { styleTags, tags } from '@lezer/highlight';
import { snippetCompletion, completeFromList } from '@codemirror/autocomplete';
import { ExternalTokenizer, LRParser } from '@lezer/lr';
import { linter } from '@codemirror/lint';

const ValueDirectives = [
    "title",
    "subtitle",
    "artist",
    "album",
    "comment"
];
const Directives = [
    "start_of_verse",
    "end_of_verse",
    "start_of_chorus",
    "end_of_chorus",
    "start_of_tab",
    "end_of_tab"
];
const snippets = [
    .../*@__PURE__*/ValueDirectives.map(name => snippetCompletion(`{${name}: #{}}`, { label: name, type: "directive" })),
    .../*@__PURE__*/Directives.map(name => snippetCompletion(`{${name}}`, { label: name, type: "directive" })),
    /*@__PURE__*/snippetCompletion("\[${}\]", { label: "chord", type: "chord" }),
    /*@__PURE__*/snippetCompletion("{define: ${1:name} frets ${2:E} ${3:A} ${4:D} ${5:G} ${6:B} ${7:E} fingers ${8:E} ${9:A} ${10:D} ${11:G} ${12:B} ${13:E}}", { label: "define", type: "directive" }),
    /*@__PURE__*/snippetCompletion(/*@__PURE__*/`
      {start_of_tab}
      E | -#{1:-}--------------------------------|
      B | ----------------------------------|
      G | ----------------------------------|
      D | ----------------------------------|
      A | ----------------------------------|
      E | ----------------------------------|
      {end_of_tab}
    `.replaceAll(/^\s*/gm, ''), { label: "tab", type: "directive" })
];

// Get a list of all the chords used in the document
function chordsUsed(state) {
    const chords = {};
    syntaxTree(state).iterate({
        enter(node) {
            if (node.type.name === "Chord") {
                const name = state.doc.sliceString(node.from, node.to);
                chords[name] = (chords[name] || 0) + 1;
            }
        }
    });
    return chords;
}
function completeChords(context) {
    const word = context.matchBefore(/[^\[]*/);
    const chords = chordsUsed(context.state);
    return {
        from: word.from,
        options: Object.keys(chords).map(chord => {
            return { label: chord, type: "chord", boost: chords[chord] };
        })
    };
}
const completeSnippets = /*@__PURE__*/completeFromList(snippets);
function chordproCompletionSource(context) {
    let node = syntaxTree(context.state).resolveInner(context.pos, -1);
    if (node.name === "[" || node.name === "Chord") {
        return completeChords(context);
    }
    else {
        return completeSnippets(context);
    }
}

// This file was generated by lezer-generator. You probably shouldn't edit it.
const Comment = 1,
  blankLine = 27,
  titleText = 28,
  authorText = 29,
  SectionName = 2,
  chordLineStart = 30,
  directiveLineStart = 31,
  Chord = 3,
  Dialect_noTitle = 0;

function readSymbolNote(input) {
    if (input.next < 65 /* Ch.A */ || input.next > 71 /* Ch.G */)
        return false;
    input.advance();
    if (input.next == 98 /* Ch.b */ || input.next == 35 /* Ch.Hash */)
        input.advance();
    return true;
}
function numeralChar(code) {
    return code == 73 /* Ch.I */ || code == 86 /* Ch.V */ || code == 105 /* Ch.i */ || code == 118 /* Ch.v */;
}
function readNumeralNote(input) {
    let off = 0;
    if (input.next == 98 /* Ch.b */ || input.next == 35 /* Ch.Hash */)
        off++;
    if (!numeralChar(input.peek(off)))
        return false;
    input.advance(off + 1);
    if (numeralChar(input.next))
        input.advance();
    if (numeralChar(input.next))
        input.advance();
    return true;
}
function readNumericNote(input) {
    let off = 0;
    if (input.next == 98 /* Ch.b */ || input.next == 35 /* Ch.Hash */)
        off++;
    let next = input.peek(off);
    if (next < 49 /* Ch.One */ || next > 55 /* Ch.Seven */)
        return false;
    input.advance(off + 1);
    return true;
}
function readSuffixPart(input, off) {
    let n = input.peek(off);
    if (n < 0)
        return off;
    if (n == 98 /* Ch.b */ || n == 35 /* Ch.Hash */ || n == 111 /* Ch.o */ || n == 120 /* Ch.x */ || n == 42 /* Ch.Plus */ || n == 45 /* Ch.Dash */ ||
        n >= 48 /* Ch.Zero */ && n <= 57 /* Ch.Nine */)
        return off + 1;
    let n1 = input.peek(off + 1), n2 = input.peek(off + 2);
    if (n == 110 /* Ch.n */ && n1 == 111 /* Ch.o */)
        return off + 2; // no
    if (n == 115 /* Ch.s */ && n1 == 117 /* Ch.u */ && n2 == 115 /* Ch.s */)
        return off + 3; // sus
    if (n == 97 /* Ch.a */ && ((n1 == 100 /* Ch.d */ && n2 == 100 /* Ch.d */) || (n1 == 117 /* Ch.u */ && n2 == 103 /* Ch.g */)))
        return off + 3; // add/aug
    if (n == 100 /* Ch.d */ && (n1 == 105 /* Ch.i */ || n1 == 111 /* Ch.o */) && n2 == 109 /* Ch.m */)
        return off + 3; // dim/dom
    if (n == 117 /* Ch.u */ && n1 == 110 /* Ch.n */ && n2 == 105 /* Ch.i */ && input.peek(off + 3) == 115 /* Ch.s */)
        return off + 4; // unis
    if (n == 116 /* Ch.t */ && n1 == 114 /* Ch.r */ && n2 == 105 /* Ch.i */) {
        for (let i = off + 3; i < off + 8; i++)
            if (input.peek(i) != "triangle".charCodeAt(i))
                return off;
        return off + 8;
    }
    // Major/Majj/Minor or some prefix of that
    if (n == 109 /* Ch.m */ || n == 77 /* Ch.M */) {
        if (n1 == 97 /* Ch.a */) {
            if (n2 == 106 /* Ch.j */) {
                let n3 = input.peek(off + 3);
                if (n3 == 106 /* Ch.j */)
                    return off + 4;
                if (n3 == 111 /* Ch.o */ && input.peek(off + 4) == 114 /* Ch.r */)
                    return off + 5;
                return off + 3;
            }
            return off + 2;
        }
        else if (n1 == 105 /* Ch.i */) {
            if (n2 == 110 /* Ch.n */) {
                if (input.peek(off + 3) == 111 /* Ch.o */ && input.peek(off + 4) == 114 /* Ch.r */)
                    return off + 5;
                return off + 3;
            }
            return off + 2;
        }
        return off + 1;
    }
    return off;
}
function readSuffixParts(input, off) {
    for (;;) {
        let next = readSuffixPart(input, off);
        if (next == off)
            return off;
        off = next;
    }
}
function readChordSuffix(input) {
    let off = 0;
    for (;;) {
        if (input.peek(off) == 40 /* Ch.ParenL */) {
            let end = readSuffixParts(input, off + 1);
            if (input.peek(end) != 41 /* Ch.ParenR */)
                break;
            off = end + 1;
        }
        else {
            let end = readSuffixParts(input, off);
            if (end == off)
                break;
            off = end;
        }
    }
    if (off)
        input.advance(off);
}
function readChord(input) {
    let parens = input.next == 40 /* Ch.ParenL */;
    if (parens)
        input.advance();
    if (readSymbolNote(input)) {
        readChordSuffix(input);
        if (input.next == 47 /* Ch.Slash */) {
            input.advance();
            readSymbolNote(input);
        }
    }
    else if (readNumeralNote(input)) {
        readChordSuffix(input);
        if (input.next == 47 /* Ch.Slash */) {
            input.advance();
            readNumeralNote(input);
        }
    }
    else if (readNumericNote(input)) {
        readChordSuffix(input);
        if (input.next == 47 /* Ch.Slash */) {
            input.advance();
            if (!readNumericNote(input) && (input.next == 98 /* Ch.b */ || input.next == 35 /* Ch.Hash */))
                input.advance();
        }
    }
    else if (input.next == 47 /* Ch.Slash */) {
        input.advance();
        readSymbolNote(input) || readNumeralNote(input) || readNumericNote(input);
    }
    else if (input.next == 124 /* Ch.Pipe */ || input.next == 45 /* Ch.Dash */ || input.next == 120 /* Ch.x */) {
        input.advance();
    }
    else {
        return false;
    }
    if (parens) {
        if (input.next == 41 /* Ch.ParenR */)
            input.advance();
        else
            return false;
    }
    return true;
}
function endOfLine(ch) {
    return ch == 10 /* Ch.Newline */ || ch == 13 /* Ch.Return */ || ch == -1 /* Ch.Eof */;
}
function wordChar(ch) {
    return ch >= 48 /* Ch.Zero */ && ch <= 57 /* Ch.Nine */ || ch >= 65 /* Ch.A */ && ch <= 90 /* Ch.Z */ || ch >= 97 /* Ch.a */ && ch <= 122 /* Ch.z */ || ch == 95 /* Ch.Underscore */;
}
function hasDirectiveName(input) {
    if (!wordChar(input.next))
        return false;
    let off = 1;
    while (wordChar(input.peek(off)))
        off++;
    if (input.peek(off) != 58 /* Ch.Colon */)
        return false;
    off++;
    while (input.peek(off) == 32 /* Ch.Space */)
        off++;
    return !endOfLine(input.peek(off));
}
let sectionHeaders = ["verse", "chorus", "bridge", "tag", "interlude", "instrumental", "inst", "intro"];
let sectionHeadersUpper = /*@__PURE__*/sectionHeaders.map(h => h.toUpperCase());
function readSectionHeader(input) {
    headers: for (let i = 0; i < sectionHeaders.length; i++) {
        let header = sectionHeaders[i], upper = sectionHeadersUpper[i];
        for (let j = 0; j < header.length; j++) {
            let next = input.peek(j);
            if (next != header.charCodeAt(j) && next != upper.charCodeAt(j))
                continue headers;
        }
        let after = input.peek(header.length);
        if (endOfLine(after) || after == 32 /* Ch.Space */ || after == 58 /* Ch.Colon */ || after == 40 /* Ch.ParenL */) {
            input.advance(header.length);
            return true;
        }
    }
    return false;
}
function skipLine(input) {
    while (!endOfLine(input.next))
        input.advance();
}
const line = /*@__PURE__*/new ExternalTokenizer((input, stack) => {
    let prev = input.peek(-1), startPos = input.pos;
    if (prev != -1 && prev != 10 /* Ch.Newline */ && prev != 13 /* Ch.Return */)
        return;
    while (input.next == 32 /* Ch.Space */ || input.next == 9 /* Ch.Tab */)
        input.advance();
    let maybeComment = input.next == 35 /* Ch.Hash */;
    if (input.next == 10 /* Ch.Newline */ || input.next == 13 /* Ch.Return */) {
        input.acceptToken(blankLine);
        return;
    }
    else if (readSectionHeader(input)) {
        input.acceptToken(SectionName);
        return;
    }
    else if (hasDirectiveName(input)) {
        input.acceptToken(directiveLineStart);
    }
    else if (!stack.dialectEnabled(Dialect_noTitle) && stack.canShift(titleText) && wordChar(input.next)) {
        skipLine(input);
        input.acceptToken(titleText);
        return;
    }
    else if (!stack.dialectEnabled(Dialect_noTitle) && stack.canShift(authorText) && wordChar(input.next)) {
        skipLine(input);
        input.acceptToken(authorText);
        return;
    }
    else if (readChord(input)) {
        // See if this is a chord-only line
        for (;;) {
            if (endOfLine(input.next)) {
                input.acceptToken(chordLineStart, startPos - input.pos);
                return;
            }
            else if (input.next == 32 /* Ch.Space */ || input.next == 9 /* Ch.Tab */) {
                input.advance();
            }
            else if (!readChord(input)) {
                return;
            }
        }
    }
    else if (maybeComment) {
        skipLine(input);
        input.acceptToken(Comment);
        return;
    }
});
const chord = /*@__PURE__*/new ExternalTokenizer(input => {
    if (readChord(input))
        input.acceptToken(Chord);
});

// This file was generated by lezer-generator. You probably shouldn't edit it.
const parser = /*@__PURE__*/LRParser.deserialize({
  version: 14,
  states: "'OQoOTOOOOOO'#Ca'#CaOvOWO'#CcO{OWO'#CiO!QOYO'#CkO!YOSO'#CjO!nOSO'#CjO#SOYO'#CpO#[O`O'#CqO!zOSO'#CsQQOTOOO#jOSOOO#oOWO,58}O#wOWO,59TOOOO,59V,59VO#|OWO,59VOOOO'#Ct'#CtO$ROSO,59UOOOP,59_,59_OOOQ'#Cu'#CuO$gOQO'#CuO$lOYO,59[OOOO,59],59]O$zOWO,59]OOOP-E6q-E6qQ%VOTOOOOOO1G.i1G.iO%^OpO1G.iO%fOpO1G.oOOOO1G.q1G.qOOOO-E6r-E6rOOOQ,59a,59aO%nOUO1G.vOOOQ-E6s-E6sOOOO1G.w1G.wOOOO'#Cb'#CbO%yOSOOO&OOWO7+$TO&TOpO7+$TOOOO7+$Z7+$ZO&YOpO7+$ZQQOTOOOOOO<<Go<<GoO&_OWO<<GoOOOO<<Gu<<GuOOOOAN=ZAN=Z",
  stateData: "&g~OPXOQWOWQO`SObTOcTOkXOnVOoRO~OlPO~PQOX[O~OX]O~OR_Oa^O~OWQO`SOb`Oc`Op^Xr^X~OWQO`SOb`Oc`OpbOrbO~ORcOqdO~OYfOfgOpeXreX~OpiO~OYkO[jO~OYlO~OamO~OWQO`SOb`Oc`Op^ar^a~ORoO~ORcOqpOpdarda~OYrOpearea~OmsO~PQOZuOqvO~OZwOqxO~ORoOpdirdi~OpyO~O[zO~OZ{O~OZ|O~O[}O~OqZ~",
  goto: "!tjPPPPPknqPPPPP||!SPPPP||P!_!g!nRZORtiWUOYiyV`TUaXXOYiyWTOYiyV`TUaUYOiyRhYSaTURnaQeVRqe",
  nodeNames: "⚠ Comment SectionName Chord Song Title Author Directive { DirectiveName : DirectiveValue } Directive Lyrics InlineChord [ ] Lyric Separator ChordLine Section SectionText",
  maxTerm: 34,
  skippedNodes: [0],
  repeatNodeCount: 3,
  tokenData: "2p~RjOX!sXY*TYZ+eZ]!s]^+j^p!spq*Tq![!s![!]+p!]!c!s!c!},o!}#O.S#O#P/O#P#Q/l#Q#R!s#R#S,o#S#T!s#T#o,o#o#p0s#p#q!s#q#r1o#r;'S!s;'S;=`)}<%l~!s~O!s~~2k]!|^bPfSZWOY!sZ]!s^![!s![!]#x!]!}!s!}#O&t#O#P'n#P#o!s#o#p&t#p#q!s#q#r(}#r;'S!s;'S;=`)}<%lO!sX$P[bPZWOY#xZ]#x^!}#x!}#O$u#O#P%d#P#o#x#o#p$u#p#q#x#q#r%|#r;'S#x;'S;=`&n<%lO#xW$zWZWOY$uZ]$u^#O$u#O#P%d#P#q$u#r;'S$u;'S;=`%v<%lO$uW%gTOY$uZ]$u^;'S$u;'S;=`%v<%lO$uW%yP;=`<%l$uP&RVbPOY%|Z]%|^!}%|#P#o%|#p;'S%|;'S;=`&h<%lO%|P&kP;=`<%l%|X&qP;=`<%l#x[&{ZfSZWOY&tZ]&t^![&t![!]$u!]#O&t#O#P'n#P#q&t#q#r(`#r;'S&t;'S;=`(Y<%lO&t['sVfSOY&tZ]&t^![&t![!]$u!];'S&t;'S;=`(Y<%lO&t[(]P;=`<%l&tS(eUfSOY(`Z](`^![(`!];'S(`;'S;=`(w<%lO(`S(zP;=`<%l(`T)UZbPfSOY(}Z](}^![(}![!]%|!]!}(}!}#P(`#P#o(}#o#p(`#p;'S(};'S;=`)w<%lO(}T)zP;=`<%l(}]*QP;=`<%l!s_*`abPfSqYZWOX!sXY*TZ]!s^p!spq*Tq![!s![!]#x!]!}!s!}#O&t#O#P'n#P#o!s#o#p&t#p#q!s#q#r(}#r;'S!s;'S;=`)}<%lO!s~+jOp~~+mPYZ+e_+y[YUbPZWOY#xZ]#x^!}#x!}#O$u#O#P%d#P#o#x#o#p$u#p#q#x#q#r%|#r;'S#x;'S;=`&n<%lO#x_,zbXQbPfSZWOY!sZ]!s^![!s![!]#x!]!c!s!c!},o!}#O&t#O#P'n#P#R!s#R#S,o#S#T!s#T#o,o#o#p&t#p#q!s#q#r(}#r;'S!s;'S;=`)}<%lO!s].]Z`PfSZWOY&tZ]&t^![&t![!]$u!]#O&t#O#P'n#P#q&t#q#r(`#r;'S&t;'S;=`(Y<%lO&t]/VVfScPOY&tZ]&t^![&t![!]$u!];'S&t;'S;=`(Y<%lO&t_/w^aQbPfSZWOY!sZ]!s^![!s![!]#x!]!}!s!}#O&t#O#P'n#P#o!s#o#p&t#p#q!s#q#r(}#r;'S!s;'S;=`)}<%lO!s]0|ZWPfSZWOY&tZ]&t^![&t![!]$u!]#O&t#O#P'n#P#q&t#q#r(`#r;'S&t;'S;=`(Y<%lO&tV1xZ[QbPfSOY(}Z](}^![(}![!]%|!]!}(}!}#P(`#P#o(}#o#p(`#p;'S(};'S;=`)w<%lO(}~2pOr~",
  tokenizers: [line, chord, 0, 1, 2, 3],
  topRules: {"Song":[0,4]},
  dialects: {noTitle: 0},
  tokenPrec: 250
});

const exampleStringLinter = /*@__PURE__*/linter(view => {
    let diagnostics = [];
    syntaxTree(view.state).cursor().iterate(node => {
        if (node.name == "String")
            diagnostics.push({
                from: node.from,
                to: node.to,
                severity: "warning",
                message: "Stings are forbidden",
                actions: [{
                        name: "Remove",
                        apply(view, from, to) { view.dispatch({ changes: { from, to } }); }
                    }]
            });
    });
    return diagnostics;
});

const ChordProLanguage = /*@__PURE__*/LRLanguage.define({
    parser: /*@__PURE__*/parser.configure({
        props: [
            /*@__PURE__*/styleTags({
                Chord: tags.keyword,
                Comment: tags.lineComment,
                DirectiveName: tags.attributeName,
                DirectiveValue: tags.attributeValue,
                Title: tags.heading1,
                Author: tags.meta,
                SectionName: tags.name,
                "Section/...": tags.heading2,
                Lyric: tags.content,
                "{ }": tags.brace,
                "[ ]": tags.squareBracket,
                ":": tags.punctuation,
                Separator: tags.separator
            })
        ],
        dialect: "noTitle"
    }),
    languageData: {
        commentTokens: { line: "#" }
    }
});
function ChordPro() {
    return new LanguageSupport(ChordProLanguage, [
        ChordProLanguage.data.of({
            autocomplete: chordproCompletionSource
        })
    ]);
}

export { ChordPro, ChordProLanguage, exampleStringLinter };
